package factory;

import java.util.*;

public class ArrayList<E>
extends AbstractList<E>
{
	/* Controllo di coerenza dell'iteratore */
	long lastMod = -1;
	ArrayListIterator ali = null;

	E[] storage;
	int head, tail, len;

	public ArrayList(int startLen)
	{
		head = tail = len = 0;
		storage = (E[])(new Object[startLen]);
	}

	public ArrayList()
	{
		this(50);
	}

	/**
	* metodi necessari a rendere concreta la classe.
	*/
	protected List<E> getConcreteInstance()
	{
		return new ArrayList();
	}

	public ListIterator<E> listIterator(int index)
	{
		return new ArrayListIterator<E>(this, index);
	}

	/**
	* Metodi sovrascritti per maggiore efficienza.
	*/

	public int size()
	{
		return len;
	}

	/**
	* Classe iteratore.
	*/
	class ArrayListIterator<E>
	implements ListIterator<E>
	{
		private long createTime;
		private ArrayList<E> list;
		int last = 0;//-1 se ultima operazione e' previous, 1 se next, 0 se nessuna
		int cursor; // posizione relativa alla head della lista dell'elemento
								// ritornato dalla prossima chiamata a next().

		ArrayListIterator(ArrayList ref, int index)
		{
			createTime = System.currentTimeMillis();
			if(index < 0 || index > ref.len)
			{
				throw new IndexOutOfBoundsException("index = " + index);
			}
			list = ref;
			cursor = index;

			System.out.println("Cursore: " + cursor + "; index = " + index + "; head = " + list.head + "; storage.length = " + list.storage.length);
		}

		ArrayListIterator(ArrayList ref)
		{
			this(ref, 0);
		}

		void checkConsistency()
		{
			if(list.lastMod > createTime && !list.ali.equals(this))
			{
				throw new ConcurrentModificationException();
			}
		}

		public boolean hasNext()
		{
			checkConsistency();
			return (cursor < list.len);
		}

		public int nextIndex()
		{
			checkConsistency();
			return cursor;
		}

		public boolean hasPrevious()
		{
			checkConsistency();
			return (cursor > 0);
		}

		public int previousIndex()
		{
			checkConsistency();
			return (cursor-1);
		}

		public E next()
		{
			checkConsistency();
			if(!hasNext())
			{
				throw new NoSuchElementException("next in posizione " + cursor);
			}
			E retval = list.storage[(cursor+list.head)%list.storage.length];
			cursor++;
			last = 1;
			return retval;
		}

		public E previous()
		{
			checkConsistency();
			if(!hasPrevious())
			{
				throw new NoSuchElementException("previous in posizione " + cursor);
			}
			E retval = list.storage[(cursor+list.head-1)%list.storage.length];
			last = -1;
			cursor--;
			return retval;
		}

		public void add(E element)
		{
			checkConsistency();
			list.lastMod = System.currentTimeMillis();
			list.ali = this;

			if(list.storage.length == list.len)
			{
				// ho riempito l'array, ne devo istanziare una nuova grossa il
				// doppio e copiare tutto.
				E[] oldStorage = list.storage;
				list.storage = (E[])(new Object[oldStorage.length*2]);
				int i = 0;
				for(i=0;i<cursor;i++)
				{
					list.storage[i] = oldStorage[(i+list.head)%oldStorage.length];
				}
				list.storage[i] = element;
				for(;i<list.len;i++)
				{
					list.storage[i+1] = oldStorage[(i+list.head)%oldStorage.length];
				}
				list.head = 0;
				list.tail = (++list.len);
			}
			else // ci stiamo
			{
				int i = 0;
				for(i=list.len;i>=cursor;i--)
				{
					list.storage[(i+1+list.head)%list.storage.length] = list.storage[(i+list.head)%list.storage.length];
				}
				list.storage[(cursor+list.head)%list.storage.length] = element;
			}
			cursor++;
			list.tail = (++list.len)%list.storage.length;
			last = 0;
		}

		public void remove()
		{
			checkConsistency();
			list.lastMod = System.currentTimeMillis();
			list.ali = this;

			switch(last)
			{
				case 0:
					throw new IllegalStateException();
					//break;
				case 1://last next()
					if(list.len == 1)//c'e' un solo elemento, rimuovo quello e compatto tuti gli indici a zero
					{
						list.storage[list.head] = null;
						list.head = list.tail = list.len = 0;
						cursor = 0;
						last = 0;
						return;
					}
					else if(cursor == 1)//sono in testa, ma ho piu' di un elemento
					{
						list.storage[list.head] = null;
						list.head = (list.head + 1)%list.storage.length;
					}
					else if(cursor == list.len)//sono in coda, ma ho piu' di un elemento
					{
						list.storage[list.tail] = null;
						list.tail = (list.tail+list.storage.length-1)%list.storage.length;
						/*
						list.tail--;
						if(list.tail<0)
						{
							list.tail = list.storage.length-1
						}
						*/
					}
					else//sono in mezzo e' ho piu' di un elemento
					{
						if(cursor > list.len/2)//sono dopo la meta' della lista
						{
							int i = 0;
							for(i=cursor+list.head;i<list.len+list.head;i++)
							{
								list.storage[(i-1)%list.storage.length] = list.storage[i%list.storage.length];
							}
							list.tail = (list.tail+list.storage.length-1)%list.storage.length;
						}
						else//sono prima della meta' della lista
						{
							int i = 0;
							for(i=list.head+cursor-1;i>list.head;i--)
							{
								list.storage[i%list.storage.length] = list.storage[(i-1)%list.storage.length];
							}
							list.head = (list.head + 1)%list.storage.length;
						}
					}
					cursor--;

					break;
				case -1://last previous()
					if(list.len == 1)//c'e' un solo elemento, rimuovo quello e compatto tuti gli indici a zero
					{
						list.storage[list.head] = null;
						list.head = list.tail = list.len = 0;
						cursor = 0;
						last = 0;
						return;
					}
					else if(cursor == 0)//togliamo la testa
					{
						list.storage[list.head] = null;
						list.head = (list.head + 1)%list.storage.length;
					}
					else if(cursor == list.len-1)//togliamo la coda
					{
						list.storage[list.tail] = null;
						list.tail = (list.tail -1 + list.storage.length)%list.storage.length;
					}
					else//siamo in mezzo
					{
						if(cursor > list.len/2)//sono dopo la meta' della lista
						{
							int i = 0;
							for(i=list.head+cursor;i<list.head+list.len-1;i++)
							{
								list.storage[i%list.storage.length] = list.storage[(i+1)%list.storage.length];
							}
							list.tail = (list.tail -1 + list.storage.length)%list.storage.length;
						}
						else//sono prima della meta' della lista
						{
							int i = 0;
							for(i=list.head+cursor;i>list.head;i--)
							{
								list.storage[i%list.storage.length] = list.storage[(i-1)%list.storage.length];
							}
							list.head = (list.head + 1)%list.storage.length;
						}
					}
					break;
				default:
					System.err.println("last ha un valore impossibile: " + last);
					throw new RuntimeException("last ha un valore impossibile: " + last);
			}
			
			list.len--;
			last = 0;
			return;
		}

		public void set(E e)
		{
			checkConsistency();
			switch(last)
			{
				case 0://ne' next, ne' previous, errore
					throw new IllegalStateException();
					//break;
				case 1://next
					list.storage[(list.head+cursor-1)%list.storage.length] = e;
					break;
				case -1://previous
					list.storage[(list.head+cursor)%list.storage.length] = e;
					break;
				default://valore impossibile
					System.err.println("last ha un valore impossibile: " + last);
					throw new RuntimeException("last ha un valore impossibile: " + last);
			}
		}
	}
}
